import json
import logging
import logging.handlers
from pathlib import Path

import falcon

from addok.config import config
from addok.core import reverse, search
from addok.helpers.text import EntityTooLarge

notfound_logger = None
query_logger = None


@config.on_load
def on_load():
    if config.LOG_NOT_FOUND:
        global notfound_logger
        notfound_logger = logging.getLogger('notfound')
        notfound_logger.setLevel(logging.DEBUG)
        filename = Path(config.LOG_DIR).joinpath('notfound.log')
        try:
            handler = logging.handlers.TimedRotatingFileHandler(
                                                str(filename), when='midnight')
        except FileNotFoundError:
            print('Unable to write to {}'.format(filename))
        else:
            notfound_logger.addHandler(handler)

    if config.LOG_QUERIES:
        global query_logger
        query_logger = logging.getLogger('queries')
        query_logger.setLevel(logging.DEBUG)
        filename = Path(config.LOG_DIR).joinpath('queries.log')
        try:
            handler = logging.handlers.TimedRotatingFileHandler(
                                                str(filename), when='midnight')
        except FileNotFoundError:
            print('Unable to write to {}'.format(filename))
        else:
            query_logger.addHandler(handler)


def log_notfound(query):
    if config.LOG_NOT_FOUND:
        notfound_logger.debug(query)


def log_query(query, results):
    if config.LOG_QUERIES:
        if results:
            result = str(results[0])
            score = str(round(results[0].score, 2))
        else:
            result = '-'
            score = '-'
        query_logger.debug('\t'.join([query, result, score]))


class CorsMiddleware:

    def process_response(self, req, resp, resource, req_succeeded):
        resp.set_header('Access-Control-Allow-Origin', '*')
        resp.set_header('Access-Control-Allow-Headers', 'X-Requested-With')


class View:

    config = config

    def match_filters(self, req):
        filters = {}
        for name in config.FILTERS:
            req.get_param(name, store=filters)
        return filters

    def to_geojson(self, req, resp, results, query=None, filters=None,
                   center=None, limit=None):
        results = {
            "type": "FeatureCollection",
            "version": "draft",
            "features": [r.format() for r in results],
            "attribution": config.ATTRIBUTION,
            "licence": config.LICENCE,
        }
        if query:
            results['query'] = query
        if filters:
            results['filters'] = filters
        if center:
            results['center'] = center
        if limit:
            results['limit'] = limit
        self.json(req, resp, results)

    def json(self, req, resp, content):
        resp.text = json.dumps(content)
        resp.content_type = 'application/json; charset=utf-8'

    def parse_lon_lat(self, req):
        try:
            lat = float(req.get_param('lat'))
            for key in ('lon', 'lng', 'long'):
                lon = req.get_param(key)
                if lon is not None:
                    lon = float(lon)
                    break
        except (ValueError, TypeError):
            lat = None
            lon = None
        return lon, lat


class Search(View):

    def on_get(self, req, resp, **kwargs):
        query = req.get_param('q')
        if not query:
            raise falcon.HTTPBadRequest(title='Missing query')
        limit = req.get_param_as_int('limit') or 5  # use config
        autocomplete = req.get_param_as_bool('autocomplete')
        if autocomplete is None:
            # Default is True.
            # https://github.com/falconry/falcon/pull/493#discussion_r44376219
            autocomplete = True
        lon, lat = self.parse_lon_lat(req)
        center = None
        if lon and lat:
            center = (lon, lat)
        filters = self.match_filters(req)
        try:
            results = search(query, limit=limit, autocomplete=autocomplete,
                             lat=lat, lon=lon, **filters)
        except EntityTooLarge as e:
            raise falcon.HTTPPayloadTooLarge(title=str(e))
        if not results:
            log_notfound(query)
        log_query(query, results)
        self.to_geojson(req, resp, results, query=query, filters=filters,
                        center=center, limit=limit)


class Reverse(View):

    def on_get(self, req, resp, **kwargs):
        lon, lat = self.parse_lon_lat(req)
        if lon is None or lat is None:
            raise falcon.HTTPBadRequest(title='Invalid args')
        limit = req.get_param_as_int('limit') or 1
        filters = self.match_filters(req)
        results = reverse(lat=lat, lon=lon, limit=limit, **filters)
        self.to_geojson(req, resp, results, filters=filters, limit=limit)


def register_http_endpoint(api):
    api.add_route('/search', Search())
    api.add_route('/reverse', Reverse())


def register_command(subparsers):
    parser = subparsers.add_parser('serve', help='Run debug server')
    parser.set_defaults(func=run)
    parser.add_argument('--host', default='127.0.0.1',
                        help='Host to expose the demo serve on')
    parser.add_argument('--port', default='7878',
                        help='Port to expose the demo server on')


def run(args):
    # Do not import at load time for preventing config import loop.
    from .wsgi import simple
    simple(args)



import os
from pathlib import Path

REDIS = {
    'host': os.environ.get('REDIS_HOST') or 'localhost',
    'port': os.environ.get('REDIS_PORT') or 6379,
    'unix_socket_path': os.environ.get('REDIS_SOCKET'),
    'indexes': {
        'db': os.environ.get('REDIS_DB_INDEXES') or 0,
    },
    'documents': {
        'db': os.environ.get('REDIS_DB_DOCUMENTS') or 1,
    }
}

# Min/max number of results to be retrieved from db and scored.
BUCKET_MIN = 10
BUCKET_MAX = 100

# Above this treshold, terms are considered commons.
COMMON_THRESHOLD = 10000

# Above this treshold, we avoid intersecting sets.
INTERSECT_LIMIT = 100000

# Min score considered matching the query.
MATCH_THRESHOLD = 0.9

# Do not consider result if final score is below this threshold.
MIN_SCORE = 0.1

QUERY_MAX_LENGTH = 200

GEOHASH_PRECISION = 7

MIN_EDGE_NGRAMS = 3
MAX_EDGE_NGRAMS = 20

SYNONYMS_PATH = None

# Pipeline stream to be used.
PROCESSORS_PYPATHS = [  # Rename in TOKEN_PROCESSORS / STRING_PROCESSORS?
    'addok.helpers.text.tokenize',
    'addok.helpers.text.normalize',
    'addok.helpers.text.flag_housenumber',
    'addok.helpers.text.synonymize',
]
QUERY_PROCESSORS_PYPATHS = [
    'addok.helpers.text.check_query_length',
]
# Remove SEARCH_PREFIXES.
SEARCH_PREPROCESSORS_PYPATHS = [
    'addok.helpers.search.tokenize',
    'addok.helpers.search.search_tokens',
    'addok.helpers.search.select_tokens',
    'addok.helpers.search.set_should_match_threshold',
]
BATCH_PROCESSORS_PYPATHS = [
    'addok.batch.to_json',
    'addok.helpers.index.prepare_housenumbers',
    'addok.ds.store_documents',
    'addok.helpers.index.index_documents',
]
BATCH_FILE_LOADER_PYPATH = 'addok.helpers.load_file'
BATCH_CHUNK_SIZE = 1000
# During imports, workers are consuming RAM;
# let one process free for Redis by default.
BATCH_WORKERS = max(os.cpu_count() - 1, 1)
RESULTS_COLLECTORS_PYPATHS = [
    'addok.helpers.collectors.no_tokens_but_housenumbers_and_geohash',
    'addok.helpers.collectors.no_available_tokens_abort',
    'addok.helpers.collectors.only_commons',
    'addok.helpers.collectors.bucket_with_meaningful',
    'addok.helpers.collectors.reduce_with_other_commons',
    'addok.helpers.collectors.ensure_geohash_results_are_included_if_center_is_given',  # noqa
    'addok.helpers.collectors.extend_results_reducing_tokens',
    'addok.helpers.collectors.extend_results_extrapoling_relations',
]
SEARCH_RESULT_PROCESSORS_PYPATHS = [
    'addok.helpers.results.match_housenumber',
    'addok.helpers.results.make_labels',
    'addok.helpers.results.score_by_importance',
    'addok.helpers.results.score_by_autocomplete_distance',
    'addok.helpers.results.score_by_ngram_distance',
    'addok.helpers.results.score_by_geo_distance',
]
REVERSE_RESULT_PROCESSORS_PYPATHS = [
    'addok.helpers.results.load_closer',
    'addok.helpers.results.make_labels',
    'addok.helpers.results.score_by_geo_distance',
]
RESULTS_FORMATTERS_PYPATHS = [
    'addok.helpers.formatters.geojson',
]
INDEXERS_PYPATHS = [
    'addok.helpers.index.HousenumbersIndexer',
    'addok.helpers.index.FieldsIndexer',
    # Pairs indexer must be after `FieldsIndexer`.
    'addok.pairs.PairsIndexer',
    # Edge ngram indexer must be after `FieldsIndexer`.
    'addok.autocomplete.EdgeNgramIndexer',
    'addok.helpers.index.FiltersIndexer',
    'addok.helpers.index.GeohashIndexer',
]
# Any object like instance having `loads` and `dumps` methods.
DOCUMENT_SERIALIZER_PYPATH = 'addok.helpers.serializers.ZlibSerializer'

DOCUMENT_STORE_PYPATH = 'addok.ds.RedisStore'

# Fields to be indexed
# If you want a housenumbers field but need to name it differently, just add
# type="housenumbers" to your field.
FIELDS = [
    {'key': 'name', 'boost': 4, 'null': False},
    {'key': 'street'},
    {'key': 'postcode',
     'boost': lambda doc: 1.2 if doc.get('type') == 'municipality' else 1},
    {'key': 'city'},
    {'key': 'housenumbers'},
    {'key': 'context'},
]

# Sometimes you only want to add some fields keeping the default ones.
EXTRA_FIELDS = []

# Weight of a document own importance:
IMPORTANCE_WEIGHT = 0.1

# Default score for the relation token => document
DEFAULT_BOOST = 1.0

# Data attribution
# Can also be an object {source: attribution}
ATTRIBUTION = "BANO"

# Data licence
# Can also be an object {source: licence}
LICENCE = "ODbL"

# Available filters (remember that every filter means bigger index)
FILTERS = ["type", "postcode"]

LOG_DIR = os.environ.get("ADDOK_LOG_DIR", Path(__file__).parent.parent.parent)

LOG_QUERIES = False
LOG_NOT_FOUND = False

INDEX_EDGE_NGRAMS = True



import redis
from hashids import Hashids

from addok.config import config


hashids = Hashids()


class RedisProxy:
    instance = None
    Error = redis.RedisError

    def connect(self, *args, **kwargs):
        self.instance = redis.Redis(*args, **kwargs)

    def __getattr__(self, name):
        return getattr(self.instance, name)

    def next_id(self):
        next_id = self.incr('_id_sequence')
        return hashids.encode(next_id)


DB = RedisProxy()


@config.on_load
def connect():
    params = config.REDIS.copy()
    params.update(config.REDIS.get('indexes', {}))
    DB.connect(
        host=params.get('host'),
        port=params.get('port'),
        db=params.get('db'),
        password=params.get('password'),
        unix_socket_path=params.get('unix_socket_path'),
    )



"""
Exposes a pytest plugin so addok plugins do not need to recode all the test
setup logic.
"""

import json
import os
from pathlib import Path
import uuid

import pytest
# Do not import files from the top of the module, otherwise they will
# not be taken into account by the coverage.


def pytest_configure():
    from addok.config import config as addok_config
    addok_config.__class__.TESTING = True
    # Force test config.
    from addok import config as config_module
    os.environ['ADDOK_CONFIG_MODULE'] = str(
        Path(config_module.__file__).parent / 'test.py')
    import logging
    logging.basicConfig(level=logging.DEBUG)
    addok_config.REDIS['indexes']['db'] = 14
    addok_config.REDIS['documents']['db'] = 15
    addok_config.load()


def pytest_runtest_setup(item):
    from addok import db, ds
    from addok.config import config as addok_config
    assert db.DB.connection_pool.connection_kwargs['db'] == 14
    if addok_config.DOCUMENT_STORE == ds.RedisStore:
        assert ds._DB.connection_pool.connection_kwargs['db'] == 15


def pytest_runtest_teardown(item, nextitem):
    from addok import db, ds
    from addok.config import config as addok_config
    assert db.DB.connection_pool.connection_kwargs['db'] == 14
    db.DB.flushdb()
    if addok_config.DOCUMENT_STORE == ds.RedisStore:
        assert ds._DB.connection_pool.connection_kwargs['db'] == 15
        ds._DB.flushdb()


def pytest_addoption(parser):
    group = parser.getgroup("general")
    group._addoption('--shell',
                     action="store_true", dest="addokshell", default=False,
                     help="start addok interactive shell on error.")


def pytest_exception_interact(node, call, report):
    if node.config.getvalue("addokshell"):
        from addok.shell import invoke
        invoke()


class DummyDoc(dict):

    def __init__(self, *args, **kwargs):
        skip_index = kwargs.pop('skip_index', False)
        super().__init__(*args, **kwargs)
        if not skip_index:
            self.index()

    def update(self, **kwargs):
        super().update(kwargs)
        self.index()

    def index(self):
        from addok.batch import process_documents
        process_documents(json.dumps(self.copy()))


@pytest.fixture
def factory(request):
    from addok import db

    def _(**kwargs):
        default = {
            'id': uuid.uuid4().hex,
            '_id': db.DB.next_id(),
            'type': 'street',
            'name': 'ellington',
            'importance': 0.0,
            'lat': '48.3254',
            'lon': '2.256'
        }
        default.update(kwargs)
        doc = DummyDoc(**default)
        return doc
    return _


@pytest.fixture
def street(factory):
    return factory()


@pytest.fixture
def city(factory):
    return factory(type='city')


@pytest.fixture
def housenumber(factory):
    return factory(housenumbers={'11': {'lat': '48.3254', 'lon': '2.256'}})


@pytest.fixture
def app():
    # Do not import before redis config has been
    # patched.
    from addok.http.wsgi import application
    return application


class MonkeyPatchWrapper(object):
    def __init__(self, monkeypatch, wrapped_object):
        super().__setattr__('monkeypatch', monkeypatch)
        super().__setattr__('wrapped_object', wrapped_object)

    def __getattr__(self, attr):
        return getattr(self.wrapped_object, attr)

    def __setattr__(self, attr, value):
        self.monkeypatch.setattr(self.wrapped_object, attr, value,
                                 raising=False)

    def __delattr__(self, attr):
        self.monkeypatch.delattr(self.wrapped_object, attr)


@pytest.fixture()
def config(request, monkeypatch):
    from addok.config import config as addok_config
    return MonkeyPatchWrapper(monkeypatch, addok_config)



from collections import OrderedDict

from pkg_resources import iter_entry_points

plugins = OrderedDict()
blocked_plugins = set([])


def load():
    for ep in iter_entry_points('addok.ext'):
        register(ep.load(), ep.name)


def register(module, name=None):
    if name is None:
        name = module.__name__
    if name in blocked_plugins:
        print('Requested registration of plugin {} but this plugin is '
              'blocked'.format(name))
        return
    plugins[name] = module


def spec(func):

    def caller(*args, **kwargs):
        for plugin in plugins.copy().values():
            try:
                getattr(plugin, func.__name__)(*args, **kwargs)
            except AttributeError:
                pass
    return caller


def block(name_or_module):
    if not isinstance(name_or_module, str):
        name_or_module = name_or_module.__name__
    if name_or_module in plugins:
        del plugins[name_or_module]
    blocked_plugins.add(name_or_module)


@spec
def register_http_endpoint(api):
    """Add new endpoints to Addok API."""


@spec
def register_http_middleware(middlewares):
    """Add new middlewares to Addok API."""


@spec
def preconfigure(config):
    """Configure addok by patching config object before user local config."""


@spec
def configure(config):
    """Configure addok by patching config object after user local config."""


@spec
def register_command(subparsers):
    """Register command for Addok CLI."""


@spec
def register_shell_command(cmd):
    """Register command for Addok shell."""



import os
import sys
from functools import wraps
from importlib import import_module
from math import asin, cos, exp, radians, sin, sqrt
from multiprocessing.pool import RUN, IMapUnorderedIterator, Pool

from progressist import ProgressBar

from addok.config import config

PYTHON_VERSION = sys.version_info


def load_file(filepath):
    with open(filepath) as f:
        for line in f:
            yield line


def load_msgpack_file(filepath):
    import msgpack  # We don't want to make it a required dependency.
    with open(filepath, mode='rb') as f:
        for line in msgpack.Unpacker(f, encoding='utf-8'):
            yield line


def iter_pipe(pipe, processors):
    """Allow for iterators to return either an item or an iterator of items."""
    if isinstance(pipe, str):
        pipe = [pipe]
    for it in processors:
        pipe = it(pipe)
    yield from pipe


def import_by_path(path):
    """
    Import functions or class by their path. Should be of the form:
    path.to.module.func
    """
    if not isinstance(path, str):
        return path
    module_path, *name = path.rsplit('.', 1)
    func = import_module(module_path)
    if name:
        func = getattr(func, name[0])
    return func


def yielder(func):
    @wraps(func)
    def wrapper(pipe):
        for item in pipe:
            yield func(item)
    return wrapper


def haversine_distance(point1, point2):
    """
    Calculate the great circle distance between two points
    on the earth (specified in decimal degrees).
    """
    lat1, lon1 = point1
    lat2, lon2 = point2

    # Convert decimal degrees to radians.
    lon1, lat1, lon2, lat2 = map(radians, [lon1, lat1, lon2, lat2])

    # Haversine formula.
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat/2)**2 + cos(lat1) * cos(lat2) * sin(dlon/2)**2
    c = 2 * asin(sqrt(a))

    # 6367 km is the radius of the Earth.
    km = 6367 * c
    return km


def km_to_score(km):
    # Score between 0 and 0.1 (close to 0 km will be close to 0.1, and 100 and
    # above will be 0).
    return 0.0 if km > 100 else .1 * exp(-(km / 50.0) ** 2)


COLORS = {
    'red': '31',
    'green': '32',
    'yellow': '33',
    'blue': '34',
    'magenta': '35',
    'cyan': '36',
    'white': '37',
    'reset': '39'
}


def colorText(s, color):
    # color should be a string from COLORS
    return '\033[%sm%s\033[%sm' % (COLORS[color], s, COLORS['reset'])


def red(s):
    return colorText(s, 'red')


def green(s):
    return colorText(s, 'green')


def yellow(s):
    return colorText(s, 'yellow')


def blue(s):
    return colorText(s, 'blue')


def magenta(s):
    return colorText(s, 'magenta')


def cyan(s):
    return colorText(s, 'cyan')


def white(s):
    return colorText(s, 'white')


class Bar(ProgressBar):
    animation = '{spinner}'
    template = '{prefix} {animation} Done: {done} | Elapsed: {elapsed}'


class ChunkedPool(Pool):

    def imap_unordered(self, func, iterable, chunksize):
        """Customized version of imap_unordered.

        Directly send chunks to func, instead of iterating in each process and
        sending one by one.

        Original:
        https://hg.python.org/cpython/file/tip/Lib/multiprocessing/pool.py#l271

        Other tried options:
        - map_async: makes a list(iterable), so it loads all the data for each
          process into RAM
        - apply_async: needs manual chunking
        """
        assert self._state == RUN
        task_batches = Pool._get_tasks(func, iterable, chunksize)
        result = IMapUnorderedIterator(self if PYTHON_VERSION >= (3, 8) else self._cache)
        tasks = ((result._job, i, func, chunk, {})
                 for i, (_, chunk) in enumerate(task_batches))
        self._taskqueue.put((tasks, result._set_length))
        return result


def parallelize(func, iterable, chunk_size, **bar_kwargs):
    bar = Bar(prefix='Processing…', **bar_kwargs)
    with ChunkedPool(processes=config.BATCH_WORKERS) as pool:
        for chunk in pool.imap_unordered(func, iterable, chunk_size):
            bar(step=len(chunk))
        bar.finish()



import os
import time

import geohash

from .config import config
from .db import DB
from .ds import get_document, get_documents
from .helpers import keys as dbkeys, scripts
from .helpers.text import ascii


def compute_geohash_key(geoh, with_neighbors=True):
    if with_neighbors:
        neighbors = geohash.expand(geoh)
        neighbors = [dbkeys.geohash_key(n) for n in neighbors]
    else:
        neighbors = [geoh]
    key = 'gx|{}'.format(geoh)
    total = DB.sunionstore(key, neighbors)
    if not total:
        # No need to keep it.
        DB.delete(key)
        key = False
    else:
        DB.expire(key, 10)
    return key


class Result:

    def __init__(self, _id):
        self.housenumber = None
        self._scores = {}
        self.load(_id)
        self.labels = []

    def load(self, doc_or_id):
        self._cache = {}
        if isinstance(doc_or_id, dict):
            doc = doc_or_id
        else:
            doc = get_document(doc_or_id)
        if not doc:
            raise ValueError('id "{}" not found'.format(doc_or_id[2:]))
        self._doc = doc

    def __getattr__(self, key):
        if key not in self._cache:
            # By convention, in case of multiple values, first value is default
            # value, others are aliases.
            value = self._rawattr(key)[0]
            self._cache[key] = value
        return self._cache[key]

    def __str__(self):
        return (str(self.labels[0]) if self.labels
                else self._rawattr(config.NAME_FIELD)[0])

    def _rawattr(self, key):
        value = self._doc.get(key, '')
        if not isinstance(value, (tuple, list)):
            value = [value]
        return value

    def __repr__(self):
        return '<{} - {} ({})>'.format(str(self), self._id, self.score)

    @property
    def keys(self):
        keys = ['housenumber'] + list(self._doc.keys())
        # housenumbers are too verbose for a given street.
        yield from (key for key in keys if key != 'housenumbers')

    def update(self, data):
        self._doc.update(data)
        self._cache.update(data)

    def format(self):
        result = self
        for formatter in config.RESULTS_FORMATTERS:
            result = formatter(result)
        return result

    def add_score(self, name, score, ceiling):
        if score >= self._scores.get(name, (0, 0))[0]:
            self._scores[name] = (score, ceiling)

    @property
    def score(self):
        if self._score != '':
            return float(self._score)
        score, _max = zip(*self._scores.values())
        return sum(score) / sum(_max)

    @score.setter
    def score(self, value):
        self._score = value

    @property
    def str_distance(self):
        return self._scores.get('str_distance', [0.0])[0]

    @classmethod
    def from_id(self, _id):
        """Return a result from it's document _id."""
        return Result(dbkeys.document_key(_id))


class BaseHelper:

    def __init__(self, verbose):
        self._start = time.time()
        self._debug = []
        if not verbose:
            self.debug = lambda *args: None

    def debug(self, *args):
        s = args[0] % args[1:]
        s = '[{}] {}'.format(str((time.time() - self._start) * 1000)[:5], s)
        self._debug.append(s)

    def report(self):  # pragma: no cover
        for msg in self._debug:
            print(msg)


class Search(BaseHelper):

    MAX_MEANINGFUL = 10

    def __init__(self, match_all=False, fuzzy=1, limit=10, autocomplete=True,
                 verbose=False):
        super().__init__(verbose=verbose)
        self.match_all = match_all
        self.fuzzy = fuzzy
        self.wanted = limit
        self.autocomplete = autocomplete
        self.pid = os.getpid()  # Unique id for tmp values in redis.

    def __call__(self, query, lat=None, lon=None, **filters):
        self.lat = lat
        self.lon = lon
        self._geohash_key = None
        self.results = {}
        self.bucket = set([])  # No duplicates.
        self.meaningful = []
        self.not_found = []
        self.common = []
        self.housenumbers = []
        self.keys = []
        self.matched_keys = set([])
        self.check_housenumber = filters.get('type') in [None, "housenumber"]
        self.filters = [dbkeys.filter_key(k, v.strip())
                        for k, v in filters.items() if v.strip()]
        self.query = ascii(query.strip())
        for func in config.SEARCH_PREPROCESSORS:
            func(self)
        self.debug('Taken tokens: %s', self.meaningful)
        self.debug('Common tokens: %s', self.common)
        self.debug('Housenumbers token: %s', self.housenumbers)
        self.debug('Not found tokens: %s', self.not_found)
        self.debug('Filters: %s', ['{}={}'.format(k, v)
                                   for k, v in filters.items()])
        for collector in config.RESULTS_COLLECTORS:
            self.debug('** %s **', collector.__name__.upper())
            if collector(self):
                break
        return self.render()

    @property
    def geohash_key(self):
        if self.lat and self.lon and self._geohash_key is None:
            geoh = geohash.encode(self.lat, self.lon, config.GEOHASH_PRECISION)
            self._geohash_key = compute_geohash_key(geoh)
            if self._geohash_key:
                self.debug('Computed geohash key %s', self._geohash_key)
            else:
                self.debug('Empty geohash key, deleting %s', self._geohash_key)
        return self._geohash_key

    def render(self):
        self.convert()
        self._sorted_bucket = list(self.results.values())
        self._sorted_bucket.sort(key=lambda r: r.score, reverse=True)
        return [r for r in self._sorted_bucket[:self.wanted]
                if r.score >= config.MIN_SCORE]

    def intersect(self, keys, limit=0):
        if not limit > 0:
            limit = config.BUCKET_MAX
        ids = []
        if keys:
            if self.filters:
                keys.extend(self.filters)
            if len(keys) == 1:
                key = keys[0]
                if key.startswith(dbkeys.TOKEN_PREFIX):
                    ids = DB.zrevrange(key, 0, limit - 1)
                else:
                    ids = DB.smembers(key)
            else:
                ids = scripts.zinter(keys=set(keys), args=[self.pid, limit])
        return set(ids)

    def add_to_bucket(self, keys, limit=None):
        self.debug('Adding to bucket with keys %s', keys)
        self.matched_keys.update([k for k in keys
                                  if k.startswith(dbkeys.TOKEN_PREFIX)])
        limit = limit or (config.BUCKET_MAX - len(self.bucket))
        self.bucket.update(self.intersect(keys, limit))
        self.debug('%s ids in bucket so far', len(self.bucket))

    def new_bucket(self, keys, limit=0):
        self.debug('New bucket with keys %s and limit %s', keys, limit)
        self.matched_keys = set([k for k in keys
                                 if k.startswith(dbkeys.TOKEN_PREFIX)])
        self.bucket = self.intersect(keys, limit)
        self.debug('%s ids in bucket so far', len(self.bucket))

    def convert(self):
        self.debug('Computing results')
        ids = [i for i in self.bucket if i not in self.results]
        if ids:
            documents = get_documents(*ids)
            self.debug('Done getting results data')
            for _id, doc in documents:
                result = Result(doc)
                for processor in config.SEARCH_RESULT_PROCESSORS:
                    processor(self, result)
                self.results[_id] = result
        self.debug('Done computing results')

    @property
    def bucket_full(self):
        l = len(self.bucket)
        return l >= self.wanted and l < config.BUCKET_MAX

    @property
    def bucket_overflow(self):
        return len(self.bucket) >= config.BUCKET_MAX

    @property
    def bucket_dry(self):
        return len(self.bucket) < self.wanted

    @property
    def bucket_empty(self):
        return not self.bucket

    @property
    def cream(self):
        return len([r for _id, r in self.results.items()
                    if r.str_distance >= config.MATCH_THRESHOLD])

    def has_cream(self):
        if (self.bucket_empty
           or self.bucket_overflow
           or len(self.bucket) > config.BUCKET_MIN):
            return False
        self.debug('Checking cream.')
        self.convert()
        return self.cream > 0

    @property
    def pass_should_match_threshold(self):
        return len(self.matched_keys) >= self.should_match_threshold

    @property
    def only_commons(self):
        return self.tokens and len(self.tokens) == len(self.common)


class Reverse(BaseHelper):

    def __call__(self, lat, lon, limit=1, **filters):
        self.lat = lat
        self.lon = lon
        self.keys = set([])
        self.results = []
        self.wanted = limit
        self.fetched = []
        self.check_housenumber = filters.get('type') in [None, "housenumber"]
        self.filters = [dbkeys.filter_key(k, v) for k, v in filters.items()]
        geoh = geohash.encode(lat, lon, config.GEOHASH_PRECISION)
        hashes = self.expand([geoh])
        self.fetch(hashes)
        if not self.keys:
            hashes = self.expand(hashes)
            self.fetch(hashes)
        return self.convert()

    def expand(self, hashes):
        new = []
        for h in hashes:
            neighbors = geohash.expand(h)
            for n in neighbors:
                if n not in self.fetched:
                    new.append(n)
        return new

    def fetch(self, hashes):
        self.debug('Fetching %s', hashes)
        for h in hashes:
            k = dbkeys.geohash_key(h)
            self.intersect(k)
            self.fetched.append(h)

    def intersect(self, key):
        if self.filters:
            keys = DB.sinter([key] + self.filters)
        else:
            keys = DB.smembers(key)
        self.keys.update(keys)

    def convert(self):
        for _id in self.keys:
            result = Result(_id)
            for processor in config.REVERSE_RESULT_PROCESSORS:
                processor(self, result)
            self.results.append(result)
            self.debug(result, result.distance, result.score)
        self.results.sort(key=lambda r: r.score, reverse=True)
        return self.results[:self.wanted]


def search(query, match_all=False, fuzzy=1, limit=10, autocomplete=False,
           lat=None, lon=None, verbose=False, **filters):
    helper = Search(match_all=match_all, fuzzy=fuzzy, limit=limit,
                    verbose=verbose, autocomplete=autocomplete)
    return helper(query, lat=lat, lon=lon, **filters)


def reverse(lat, lon, limit=1, verbose=False, **filters):
    helper = Reverse(verbose=verbose)
    return helper(lat, lon, limit, **filters)



import json
import os.path
import sys
from datetime import timedelta

from addok.config import config
from addok.db import DB
from addok.ds import DS
from addok.helpers import iter_pipe, parallelize, yielder


def run(args):
    if args.filepath:
        for path in args.filepath:
            process_file(path)
    elif not sys.stdin.isatty():  # Any better way to check for stdin?
        process_stdin(sys.stdin)


def reset(args):
    if args.force or input('Type "yes" to delete ALL data: ') == 'yes':
        DB.flushdb()
        DS.flushdb()
        print('All data has been deleted.')
    else:
        print('Nothing has been deleted.')


def register_command(subparsers):
    parser = subparsers.add_parser('batch', help='Batch import documents')
    parser.add_argument('filepath', nargs='*',
                        help='Path to file to process')
    parser.set_defaults(func=run)
    parser = subparsers.add_parser('reset',
                                   help='Delete ALL indexes and documents')
    parser.add_argument('--force', help='Do not ask for confirm',
                        action='store_true')
    parser.set_defaults(func=reset)


def process_file(filepath):
    print('Import from file', filepath)
    _, ext = os.path.splitext(filepath)
    if not os.path.exists(filepath):
        sys.stderr.write('File not found: {}'.format(filepath))
        sys.exit(1)
    config.INDEX_EDGE_NGRAMS = False  # Run command "ngrams" instead.
    batch(config.BATCH_FILE_LOADER(filepath))


def process_stdin(stdin):
    print('Import from stdin')
    batch(stdin)


@yielder
def to_json(row):
    try:
        return json.loads(row)
    except ValueError:
        return None


def process_documents(*docs):
    return list(iter_pipe(docs, config.BATCH_PROCESSORS))


def batch(iterable):
    parallelize(process_documents, iterable,
                chunk_size=config.BATCH_CHUNK_SIZE,
                throttle=timedelta(seconds=1))



import logging
import logging.handlers
from pathlib import Path

from addok.config import config


notfound_logger = None
query_logger = None
slow_query_logger = None


def get_logger(name):
    logger = logging.getLogger(name)
    logger.setLevel(logging.DEBUG)
    filename = Path(config.LOG_DIR).joinpath("{}.log".format(name))
    try:
        handler = logging.handlers.TimedRotatingFileHandler(
            str(filename), when="midnight"
        )
    except FileNotFoundError:
        print("Unable to write to {}".format(filename))
    else:
        logger.addHandler(handler)
    return logger


@config.on_load
def on_load():
    if config.LOG_NOT_FOUND:
        global notfound_logger
        notfound_logger = get_logger("notfound")

    if config.LOG_QUERIES:
        global query_logger
        query_logger = get_logger("queries")

    if config.SLOW_QUERIES:
        global slow_query_logger
        slow_query_logger = get_logger("slow_queries")


def log_notfound(query):
    if config.LOG_NOT_FOUND:
        notfound_logger.debug(query)


def log_query(query, results):
    if config.LOG_QUERIES:
        if results:
            result = str(results[0])
            score = str(round(results[0].score, 2))
        else:
            result = "-"
            score = "-"
        query_logger.debug("\t".join([query, result, score]))


def log_slow_query(query, results, timer):
    if config.SLOW_QUERIES:
        if results:
            result = str(results[0])
            score = str(round(results[0].score, 2))
            id_ = results[0].id
        else:
            result = "-"
            score = "-"
            id_ = "-"
        slow_query_logger.debug("\t".join([str(timer), query, id_, result, score]))



import json

import pytest

from addok.batch import process_documents
from addok.core import search


@pytest.mark.parametrize("input,expected", [
    ('rue du 8 mai troyes', False),
    ('8 rue du 8 mai troyes', '8'),
    ('3 rue du 8 mai troyes', '3'),
])
def test_match_housenumber(input, expected):
    doc = {
        'id': 'xxxx',
        '_id': 'yyyy',
        'type': 'street',
        'name': 'rue du 8 Mai',
        'city': 'Troyes',
        'lat': '49.32545',
        'lon': '4.2565',
        'housenumbers': {
            '3': {
                'lat': '48.325451',
                'lon': '2.25651'
            },
            '3 bis': {
                'lat': '48.325451',
                'lon': '2.25651'
            },
            '8': {
                'lat': '48.325451',
                'lon': '2.25651'
            },
        }
    }
    process_documents(json.dumps(doc))
    result = search(input)[0]
    assert (result.type == 'housenumber') == bool(expected)
    if expected:
        assert result.housenumber == expected



ATTRIBUTION = 'BAN'
LICENCE = 'ODbL 1.0'
LOG_QUERIES = True
LOG_NOT_FOUND = True
LOG_DIR = '/var/log/addok/'
EXTRA_FIELDS = [
    {'key': 'citycode'},
]
FILTERS = ["type", "postcode", "citycode"]

REDIS = {
    'host': '127.0.0.1',
    'port': 6379,
    'db': 0
}
QUERY_PROCESSORS_PYPATHS = [
    'addok.helpers.text.check_query_length',
    "addok_france.extract_address",
    "addok_france.clean_query",
    "addok_france.remove_leading_zeros",
]
SEARCH_RESULT_PROCESSORS_PYPATHS = [
    "addok.helpers.results.match_housenumber",
    "addok_france.make_labels",
    "addok.helpers.results.score_by_importance",
    "addok.helpers.results.score_by_autocomplete_distance",
    "addok.helpers.results.score_by_str_distance",
    "addok.helpers.results.score_by_geo_distance",
    "addok.helpers.results.adjust_scores",
]
PROCESSORS_PYPATHS = [
    "addok.helpers.text.tokenize",
    "addok.helpers.text.normalize",
    "addok_france.glue_ordinal",
    "addok_france.fold_ordinal",
    "addok_france.flag_housenumber",
    "addok.helpers.text.synonymize",
    "addok_fr.phonemicize",
]
SQLITE_DB_PATH = '/srv/addok/addok.db'



# COMMON_THRESHOLD = 10000
# INTERSECT_LIMIT = 100
LOG_QUERIES = True
LOG_NOT_FOUND = True
SLOW_QUERIES = 100
EXTRA_FIELDS = [
    {"key": "citycode"},
]
FILTERS = ["type", "postcode", "citycode", "city"]
BLOCKED_PLUGINS = ["trigrams"]
QUERY_PROCESSORS_PYPATHS = [
    "addok.helpers.text.check_query_length",
    "addok_france.extract_address",
    "addok_france.clean_query",
    "addok_france.remove_leading_zeros",
]
SEARCH_RESULT_PROCESSORS_PYPATHS = [
    "addok.helpers.results.match_housenumber",
    "addok_france.make_labels",
    "addok.helpers.results.score_by_importance",
    "addok.helpers.results.score_by_autocomplete_distance",
    "addok.helpers.results.score_by_ngram_distance",
    "addok.helpers.results.score_by_geo_distance",
    "addok.helpers.results.adjust_scores",
]
PROCESSORS_PYPATHS = [
    "addok.helpers.text.tokenize",
    "addok.helpers.text.normalize",
    "addok_france.glue_ordinal",
    "addok_france.fold_ordinal",
    "addok_france.flag_housenumber",
    "addok.helpers.text.synonymize",
    "addok_fr.phonemicize",
]
# DOCUMENT_SERIALIZER_PYPATH = 'addok.helpers.serializers.MsgpackSerializer'
# DOCUMENT_STORE_PYPATH = 'addok.ds.RedisStore'
BATCH_CHUNK_SIZE = 1000
SQLITE_DB_PATH = "/home/ybon/Code/py/addok/addok.db"
# SQLITE_DB_PATH = '/run/media/ybon/EMTEC/addok.db'
# LOG_DIR = '/pouet'
PG_CONFIG = "dbname=addok"
MIN_EDGE_NGRAMS = 2


def target_city(helper):
    if len(helper.query) <= 4:
        helper.filters.append("f|type|municipality")


RESULTS_COLLECTORS_PYPATHS = [
    target_city,
    "addok.autocomplete.only_commons_but_geohash_try_autocomplete_collector",
    "addok.helpers.collectors.no_tokens_but_housenumbers_and_geohash",
    "addok.helpers.collectors.no_available_tokens_abort",
    "addok.helpers.collectors.only_commons",
    "addok.autocomplete.no_meaningful_but_common_try_autocomplete_collector",
    "addok.autocomplete.only_commons_try_autocomplete_collector",
    "addok.helpers.collectors.bucket_with_meaningful",
    "addok.helpers.collectors.reduce_with_other_commons",
    "addok.helpers.collectors.ensure_geohash_results_are_included_if_center_is_given",  # noqa
    "addok.fuzzy.fuzzy_collector",
    "addok.autocomplete.autocomplete_meaningful_collector",
    "addok.helpers.collectors.extend_results_extrapoling_relations",
    "addok.helpers.collectors.extend_results_reducing_tokens",
]



COMMON_THRESHOLD = 500
LOG_QUERIES = False
LOG_NOT_FOUND = False
FIELDS = [
    {'key': 'name', 'boost': 4, 'null': False},
    {'key': 'code'},
    {'key': 'postal',
     'boost': lambda doc: 1.2 if doc.get('level') == 'fr/town' else 1},
]
FILTERS = ["type", "id"]
BLOCKED_PLUGINS = ['addok.pairs', 'addok.fuzzy', 'addok.autocomplete']
QUERY_PROCESSORS = []
HOUSENUMBER_PROCESSORS = []
SEARCH_RESULT_PROCESSORS = [
    'addok.helpers.results.make_labels',
    'addok.helpers.results.score_by_importance',
    'addok.helpers.results.score_by_autocomplete_distance',
    'addok.helpers.results.score_by_ngram_distance',
    'addok.helpers.results.score_by_geo_distance',
]
PROCESSORS = [
    'addok.helpers.text.tokenize',
    'addok.helpers.text.normalize',
    'addok.helpers.text.synonymize',
    'addok_trigrams.trigramize',
]
RESULTS_COLLECTORS = [
    'addok.helpers.collectors.only_commons',
    'addok.helpers.collectors.bucket_with_meaningful',
    'addok.helpers.collectors.reduce_with_other_commons',
    'addok.helpers.collectors.ensure_geohash_results_are_included_if_center_is_given',  # noqa
    'addok_trigrams.extend_results_removing_numbers',
    'addok_trigrams.extend_results_removing_one_whole_word',
    'addok_trigrams.extend_results_removing_successive_trigrams',
]
BATCH_PROCESSORS = [
    'addok_geozones.utils.prepare_document',
]



from pathlib import Path

COMMON_THRESHOLD = 500
LOG_QUERIES = False
LOG_NOT_FOUND = False
FIELDS = [
    {'key': 'STREET', 'boost': 4, 'null': False, 'name': True},
    {'key': 'POSTCODE'},
    {'key': 'UNIT'},
    {'key': 'CITY'},
    {'key': 'DISTRICT'},
    {'key': 'REGION'},
    {'key': 'housenumbers'},
]
HOUSENUMBERS_PAYLOAD_FIELDS = ['id']
FILTERS = ["type", "id"]
BLOCKED_PLUGINS = ['addok.pairs', 'addok.fuzzy', 'addok.autocomplete']
QUERY_PROCESSORS = []
HOUSENUMBER_PROCESSORS = []
PROCESSORS = [
    'addok.helpers.text.tokenize',
    'addok.helpers.text.normalize',
    'addok.helpers.text.synonymize',
    'addok_trigrams.trigramize',
]
RESULTS_COLLECTORS = [
    'addok.helpers.collectors.only_commons',
    'addok.helpers.collectors.bucket_with_meaningful',
    'addok.helpers.collectors.reduce_with_other_commons',
    'addok.helpers.collectors.ensure_geohash_results_are_included_if_center_is_given',  # noqa
    'addok_trigrams.extend_results_removing_numbers',
    'addok_trigrams.extend_results_removing_one_whole_word',
    'addok_trigrams.extend_results_removing_successive_trigrams',
]
BATCH_PROCESSORS = [
    'addok_openaddresses.utils.group_addresses',
]
SEARCH_RESULT_PROCESSORS = [
    'addok.helpers.results.match_housenumber',
    'addok_openaddress.utils.make_labels',
    'addok.helpers.results.score_by_importance',
    'addok.helpers.results.score_by_autocomplete_distance',
    'addok.helpers.results.score_by_ngram_distance',
    'addok.helpers.results.score_by_geo_distance',
]
SYNONYMS_FILENAME = 'en.txt'
OPENADDRESS_EXTRA = {
    'CITY': ['NY', 'New York']
}



ATTRIBUTION = 'BAN'
LICENCE = 'ODbL 1.0'
LOG_QUERIES = True
LOG_NOT_FOUND = True
EXTRA_FIELDS = [
    {'key': 'citycode'},
]
FILTERS = ["type", "postcode", "citycode"]
HOUSENUMBERS_PAYLOAD_FIELDS = ['id']



def geojson(result):
    properties = {
        "label": str(result),
    }
    if result._scores:
        properties["score"] = result.score
    for key in result.keys:
        val = getattr(result, key, None)
        if val and key not in ['lat', 'lon', '_id']:
            properties[key] = val
    housenumber = getattr(result, 'housenumber', None)
    if housenumber:
        if result._doc.get('type'):
            properties[result._doc['type']] = properties.get('name')
        properties['name'] = '{} {}'.format(housenumber,
                                            properties.get('name'))
    try:
        properties['distance'] = int(result.distance)
    except ValueError:
        pass
    return {
        "type": "Feature",
        "geometry": {
            "type": "Point",
            "coordinates": [float(result.lon), float(result.lat)]
        },
        "properties": properties,
    }



from datetime import date
from hashlib import md5
from io import StringIO
from pathlib import Path
from string import Template as BaseTemplate

from invoke import task


class Template(BaseTemplate):
    # Default delimiter ($) clashes at least with Nginx DSL.
    delimiter = '$$'


def render_template(path, **context):
    with Path(path).open() as f:
        template = Template(f.read())
        return StringIO(template.substitute(**context))


def as_user(ctx, user, cmd, *args, **kwargs):
    ctx.run('sudo --set-home --preserve-env --user {} '
            '{}'.format(user, cmd), *args, **kwargs)


def as_addok(ctx, cmd, *args, **kwargs):
    as_user(ctx, 'addok', cmd, *args, **kwargs)


def sudo_put(ctx, local, remote, chown=None):
    tmp = str(Path('/tmp') / md5(remote.encode()).hexdigest())
    ctx.put(local, tmp)
    ctx.run('sudo mv {} {}'.format(tmp, remote))
    if chown:
        ctx.run('sudo chown {} {}'.format(chown, remote))


@task
def addok(ctx, cmd):
    as_addok(ctx, '/srv/addok/venv/bin/addok {}'.format(cmd))


@task
def system(ctx):
    ctx.run('sudo apt update')
    ctx.run('sudo apt install redis-server python3 python3-dev '
            'python-virtualenv build-essential git wget uwsgi '
            'uwsgi-plugin-python3 bzip2 --yes')
    if not ctx.config.get('skip_nginx'):
        ctx.run('sudo apt install nginx --yes')
    ctx.run('sudo mkdir -p /etc/addok')
    ctx.run('sudo mkdir -p /var/log/addok')
    ctx.run('sudo useradd -N addok -m -d /srv/addok/ || exit 0')
    ctx.run('sudo chown addok:users /var/log/addok')
    ctx.run('sudo chsh -s /bin/bash addok')


@task
def venv(ctx):
    as_addok(ctx, 'virtualenv /srv/addok/venv --python=python3')
    as_addok(ctx, '/srv/addok/venv/bin/pip install pip -U')


@task
def settings(ctx):
    if ctx.settings:
        sudo_put(ctx, ctx.settings, '/etc/addok/addok.conf',
                 chown='addok:users')


@task
def http(ctx):
    sudo_put(ctx, 'fabfile/uwsgi_params', '/srv/addok/uwsgi_params')
    connections = ctx.config.get('connections', 256)
    socket = ctx.config.get('socket', '/srv/addok/uwsgi.sock')
    processes = ctx.config.get('processes', 4)
    uwsgi_conf = render_template('fabfile/uwsgi.ini', connections=connections,
                                 processes=processes, socket=socket)
    sudo_put(ctx, uwsgi_conf, '/etc/uwsgi/apps-enabled/addok.ini')
    if not ctx.config.get('skip_nginx'):
        nginx_conf = render_template('fabfile/nginx.conf',
                                     domain=ctx.config.domain)
        sudo_put(ctx, nginx_conf, '/etc/nginx/sites-enabled/addok')
    # On LXC containers, somaxconn cannot be changed. This must be done on the
    # host machine.
    ctx.run(
        'sudo sysctl -w net.core.somaxconn={} || exit 0'.format(connections))
    restart(ctx)


@task
def bootstrap(ctx):
    system(ctx)
    venv(ctx)
    settings(ctx)
    http(ctx)


@task
def fetch(ctx):
    ctx.run('wget {} --output-document=/tmp/data.bz2 --quiet'.format(
        ctx.config.data_uri))
    ctx.run('bunzip2 /tmp/data.bz2 --stdout > /tmp/data.json')


@task
def batch(ctx):
    ctx.run('redis-cli config set save ""')
    addok(ctx, 'batch /tmp/data.json')
    addok(ctx, 'ngrams')
    ctx.run('redis-cli save')


@task
def reload(ctx):
    fetch(ctx)
    ctx.run('sudo systemctl stop uwsgi')
    addok(ctx, 'reset')
    batch(ctx)
    restart(ctx)


@task
def deploy(ctx):
    cmd = '/srv/addok/venv/bin/pip install {} --upgrade'
    as_addok(ctx, cmd.format('addok'))
    if ctx.config.get('plugins'):
        as_addok(ctx, cmd.format(' '.join(ctx.config.plugins)))
    restart(ctx)


@task
def restart(ctx):
    ctx.run('sudo systemctl restart uwsgi')
    if not ctx.config.get('skip_nginx'):
        ctx.run('sudo systemctl restart nginx')


@task
def backup(ctx):
    today = date.today().isoformat()
    backup_filename = 'addok-backup.{}.tar.bz2'.format(today)
    ctx.run('redis-cli save')
    ctx.run('cp /etc/addok/addok.conf /tmp/local.{}.py'.format(today))
    ctx.run('cp /srv/addok/addok.db /tmp/sqlite.{}.db'.format(today))
    ctx.run('cp /var/lib/redis/dump.rdb /tmp/redis.{}.rdb'.format(today))
    ctx.run('cd /tmp && tar -jcvf {filename} sqlite.{today}.db '
            'redis.{today}.rdb local.{today}.py'.format(
                filename=backup_filename, today=today))
    scp = 'rsync --progress -e ssh'  # Allows to display progress.
    ctx.local('{scp} {ctx.user}@{ctx.host}:/tmp/{filename} .'.format(
        scp=scp, ctx=ctx, filename=backup_filename))


@task
def use_backup(ctx, backup_date=date.today().isoformat()):
    ctx.local('tar -xvjf addok-backup.{}.tar.bz2'.format(backup_date))
    with open('local.{}.py'.format(backup_date), 'r+') as configuration:
        content = configuration.read()
        if "SQLITE_DB_PATH = '/srv/addok/addok.db'" in content:
            content = content.replace(
                "SQLITE_DB_PATH = '/srv/addok/addok.db'",
                "SQLITE_DB_PATH = 'sqlite.{}.db'".format(backup_date))
            configuration.seek(0)  # Overwrite file.
            configuration.write(content)
    ctx.local('echo "dbfilename redis.{}.rdb" | redis-server -'.format(
        backup_date))



TOKEN_PREFIX = 'w|'


def token_key(s):
    return '{}{}'.format(TOKEN_PREFIX, s)


def document_key(s):
    return 'd|{}'.format(s)


def geohash_key(s):
    return 'g|{}'.format(s)


def filter_key(k, v):
    return 'f|{}|{}'.format(k, v)



import os
import imp
import importlib
import sys
try:
    import pkg_resources
except ImportError:  # pragma: no cover
    pkg_resources = None

from addok import hooks, VERSION
from . import default


class Config(dict):

    TESTING = False

    def __init__(self):
        self._post_load_func = []
        self.loaded = False
        self.plugins = [
            'addok.shell',
            'addok.http.base',
            'addok.batch',
            'addok.pairs',
            'addok.fuzzy',
            'addok.autocomplete',
        ]
        super().__init__()
        self.extend_from_object(default)

    def load(self):
        if self.loaded:
            return
        self.loaded = True
        self.load_core_plugins()
        if not Config.TESTING:
            # We don't want to autoload installed plugin during tests.
            hooks.load()  # pragma: no cover
        hooks.preconfigure(self)
        self.load_local()
        self.load_from_env()
        hooks.configure(self)
        self.resolve()
        self.post_process()
        plugins = []
        for name, plugin in hooks.plugins.items():  # pragma: no cover
            if name.startswith('addok.'):
                version = VERSION  # Core plugin.
            elif pkg_resources:
                version = pkg_resources.get_distribution(
                                plugin.__package__).version
            else:
                version = '?'
            plugins.append('{}=={}'.format(name, version))
        print('Loaded plugins:\n{}'.format(', '.join(plugins)))

    def on_load(self, func):
        self._post_load_func.append(func)
        return func

    def extend_from_object(self, obj):
        for key in dir(obj):
            if key.isupper():
                self[key] = getattr(obj, key)

    def load_core_plugins(self):
        for path in self.plugins:
            plugin = importlib.import_module(path)
            hooks.register(plugin)

    def load_local(self):
        path = (os.environ.get('ADDOK_CONFIG_MODULE')
                or os.path.join('/etc', 'addok', 'addok.conf'))
        if not os.path.exists(path):
            print('No local config file found in "{}".'.format(path))
            return

        d = imp.new_module('config')
        d.__file__ = path
        try:
            with open(path) as config_file:
                exec(compile(config_file.read(), path, 'exec'), d.__dict__)
        except (IOError, SyntaxError) as e:
            from addok.helpers import red
            print(red('Unable to import {} from '
                      'ADDOK_CONFIG_MODULE'.format(path)))
            sys.exit(e)
        else:
            print('Loaded local config from', path)
            self.extend_from_object(d)

    def load_from_env(self):
        for key, value in self.items():
            if key.isupper():
                env_key = "ADDOK_" + key
                typ = type(value)
                if typ in (list, tuple, set):
                    real_type, typ = typ, lambda x: real_type(x.split(","))
                if env_key in os.environ:
                    self[key] = typ(os.environ[env_key])

    def __getattr__(self, key):
        return self.get(key)

    def __setattr__(self, key, value):
        self[key] = value

    def post_process(self):
        self.FIELDS.extend(self.EXTRA_FIELDS)
        for field in self.FIELDS:
            key = field['key']
            if (field.get('type') == 'housenumbers' or
                    key == 'housenumbers'):
                self.HOUSENUMBERS_FIELD = key
                field['type'] = 'housenumbers'
            elif field.get('type') == 'name' or key == 'name':
                self.NAME_FIELD = key
                field['type'] = 'name'
        for func in self._post_load_func:
            func()

    def resolve(self):
        for key in list(self.keys()):
            if key.endswith('_PYPATHS'):
                self.resolve_paths(key)
            elif key.endswith('_PYPATH'):
                self.resolve_path(key)

    def resolve_path(self, key):
        from addok.helpers import import_by_path
        self[key[:-len('_PYPATH')]] = import_by_path(self[key])

    def resolve_paths(self, key):
        from addok.helpers import import_by_path
        self[key[:-len('_PYPATHS')]] = [import_by_path(path)
                                        for path in self[key]]


config = Config()



"Search engine for address. Only address."
import glob
import sys
from codecs import open  # To use a consistent encoding
from os import path

from setuptools import Extension, find_packages, setup

here = path.abspath(path.dirname(__file__))

# Get the long description from the relevant file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()


def is_pkg(line):
    return line and not line.startswith(('--', 'git', '#'))


def list_modules(dirname):
    filenames = glob.glob(path.join(dirname, '*.py'))

    module_names = []
    for name in filenames:
        module, ext = path.splitext(path.basename(name))
        if module != '__init__':
            module_names.append(module)

    return module_names


with open('requirements.txt', encoding='utf-8') as reqs:
    install_requires = [l for l in reqs.read().split('\n') if is_pkg(l)]

if sys.platform == 'darwin':
    install_requires.append('gnureadline==8.1.2')

try:
    from Cython.Distutils import build_ext
    CYTHON = True
except ImportError:
    sys.stdout.write('\nNOTE: Cython not installed. Addok will '
                     'still work fine, but may run a bit slower.\n\n')
    CYTHON = False
    cmdclass = {}
    ext_modules = []
else:
    ext_modules = [
        Extension('addok.' + ext, [path.join('addok', ext + '.py')])
        for ext in list_modules(path.join(here, 'addok'))]

    ext_modules.extend([
        Extension('addok.helpers.' + ext,
                  [path.join('addok', 'helpers', ext + '.py')])
        for ext in list_modules(path.join(here, 'addok', 'helpers'))])

    cmdclass = {'build_ext': build_ext}

VERSION = (1, 0, 4)

__author__ = 'Yohan Boniface'
__contact__ = "yohan.boniface@data.gouv.fr"
__homepage__ = "https://github.com/addok/addok"
__version__ = ".".join(map(str, VERSION))

setup(
    name='addok',
    version=__version__,
    description=__doc__,
    long_description=long_description,
    url=__homepage__,
    author=__author__,
    author_email=__contact__,
    license='WTFPL',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Topic :: Scientific/Engineering :: GIS',

        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='address openstreetmap geocoding',
    packages=find_packages(exclude=['tests']),
    install_requires=install_requires,
    extras_require={'test': ['pytest'], 'docs': 'mkdocs'},
    include_package_data=True,
    ext_modules=ext_modules,
    entry_points={
        'console_scripts': ['addok=addok.bin:main'],
        'pytest11': ['addok=addok.pytest'],
    },
    cmdclass=cmdclass,
)



#!/usr/bin/env python

import argparse
import os

from addok import VERSION
from addok.config import config


def main():
    if VERSION:
        print('Addok', VERSION)
    main_parser = argparse.ArgumentParser(description='Addok command line.',
                                          add_help=False)
    main_parser.add_argument('--config', help='Local config')
    args, extras = main_parser.parse_known_args()
    if args.config:
        os.environ['ADDOK_CONFIG_MODULE'] = args.config
    main_parser.add_argument('-h', '--help', action='store_true',
                             help='Show this help message and exit')

    subparsers = main_parser.add_subparsers(title='Available commands',
                                            metavar='')

    from addok import hooks
    config.load()
    hooks.register_command(subparsers)
    args = main_parser.parse_args(args=extras)
    if getattr(args, 'func', None):
        args.func(args)
    else:
        main_parser.print_help()


if __name__ == '__main__':
    main()



from addok.config.default import (PROCESSORS_PYPATHS, INDEXERS_PYPATHS,
                                  RESULTS_COLLECTORS_PYPATHS)
from addok.config.local import *  # noqa

RESULTS_COLLECTORS_PYPATHS.remove('addok.helpers.collectors.extend_results_reducing_tokens')  # noqa
RESULTS_COLLECTORS_PYPATHS += [
    'addok_trigrams.extend_results_removing_numbers',
    'addok_trigrams.extend_results_removing_one_whole_word',
    'addok_trigrams.extend_results_removing_successive_trigrams',
]
PROCESSORS_PYPATHS += [
    'addok_trigrams.trigramize',
]
INDEXERS_PYPATHS.remove('addok.pairs.PairsIndexer')
INDEXERS_PYPATHS.remove('addok.autocomplete.EdgeNgramIndexer')

REDIS = {
    'host': 'localhost',
    'port': 6379,
    'db': 3
}



from addok.config import config
from addok.helpers import haversine_distance, km_to_score
from addok.helpers.text import (ascii, compare_ngrams, contains, equals,
                                startswith)


def make_labels(helper, result):
    if not result.labels:
        # Make your own for better scoring (see addok-france for inspiration).
        result.labels = result._rawattr(config.NAME_FIELD)[:]
        label = result.labels[0]
        city = getattr(result, 'city', None)
        if city and city != label:
            postcode = getattr(result, 'postcode', None)
            if postcode:
                label = '{} {}'.format(label, postcode)
            label = '{} {}'.format(label, city)
            result.labels.insert(0, label)
        housenumber = getattr(result, 'housenumber', None)
        if housenumber:
            label = '{} {}'.format(housenumber, label)
            result.labels.insert(0, label)


def match_housenumber(helper, result):
    _match_housenumber(helper, result, helper.tokens)


def _match_housenumber(helper, result, tokens):
    if not helper.check_housenumber:
        return
    # Housenumber may have multiple tokens (eg. "dix huit"), we join
    # those to match the way they have been processed by
    # addok.helpers.index.prepare_housenumbers.
    raw = ''.join(sorted(helper.housenumbers, key=lambda t: t.position))
    if raw and raw in result.housenumbers:
        data = result.housenumbers[str(raw)]
        result.housenumber = data.pop('raw')
        result.type = 'housenumber'
        result.update(data)


def score_by_importance(helper, result):
    importance = getattr(result, 'importance', None)
    importance = importance or 0.0
    result.add_score('importance',
                     float(importance) * config.IMPORTANCE_WEIGHT,
                     config.IMPORTANCE_WEIGHT)


def score_by_autocomplete_distance(helper, result):
    if not helper.autocomplete:
        return
    score = 0
    query = ascii(helper.query)
    for idx, label in enumerate(result.labels):
        label = ascii(label)
        result.labels[idx] = label  # Cache ascii folding.
        if equals(query, label):
            score = 1.0
        elif startswith(query, label):
            score = 0.9
        elif contains(query, label):
            score = 0.7
        if score:
            result.add_score('str_distance', score, ceiling=1.0)
            if score >= config.MATCH_THRESHOLD:
                break
    if not score:
        _score_by_ngram_distance(helper, result, scale=0.9)


def _score_by_ngram_distance(helper, result, scale=1.0):
    for label in result.labels:
        label = ascii(label)
        score = compare_ngrams(label, helper.query) * scale
        result.add_score('str_distance', score, ceiling=1.0)
        if score >= config.MATCH_THRESHOLD:
            break


def score_by_ngram_distance(helper, result):
    if helper.autocomplete:
        return
    _score_by_ngram_distance(helper, result)


def score_by_geo_distance(helper, result):
    if helper.lat is None or helper.lon is None:
        return
    km = haversine_distance((float(result.lat), float(result.lon)),
                            (helper.lat, helper.lon))
    result.distance = km * 1000
    result.add_score('geo_distance', km_to_score(km), ceiling=0.1)


def load_closer(helper, result):
    if not helper.check_housenumber:
        return

    def sort(h):
        return haversine_distance((float(h['lat']), float(h['lon'])),
                                  (helper.lat, helper.lon))

    candidates = []
    if result.housenumbers:
        candidates = list(result.housenumbers.values())
    candidates.append({'raw': None, 'lat': result.lat, 'lon': result.lon})
    candidates.sort(key=sort)
    closer = candidates[0]
    if closer['raw']:  # Means a housenumber is closer than street center.
        result.housenumber = closer.pop('raw')
        result.type = 'housenumber'
        result.update(closer)
